@@ -47,8 +47,9 @@ module Puma
       @original_argv = @argv.dup
       @config        = conf
 
-      @binder        = Binder.new(@events)
-      @binder.import_from_env
+      @binder        = Binder.new(@events, conf)
+      @binder.create_inherited_fds(ENV).each { |k| ENV.delete k }
+      @binder.create_activated_fds(ENV).each { |k| ENV.delete k }
 
       @environment = conf.environment
 
@@ -57,6 +58,13 @@ module Puma
 
       @config.load
 
+      if @config.options[:bind_to_activated_sockets]
+        @config.options[:binds] = @binder.synthesize_binds_from_activated_fs(
+          @config.options[:binds],
+          @config.options[:bind_to_activated_sockets] == 'only'
+        )
+      end
+
       @options = @config.options
       @config.clamp
 
@@ -69,10 +77,6 @@ module Puma
         unsupported "worker mode not supported on #{RUBY_ENGINE} on this platform"
       end
 
-      if @options[:daemon] && Puma.windows?
-        unsupported 'daemon mode not supported on Windows'
-      end
-
       Dir.chdir(@restart_dir)
 
       prune_bundler if prune_bundler?
@@ -90,6 +94,8 @@ module Puma
       Puma.stats_object = @runner
 
       @status = :run
+
+      log_config if ENV['PUMA_LOG_CONFIG']
     end
 
     attr_reader :binder, :events, :config, :options, :restart_dir
@@ -105,6 +111,7 @@ module Puma
       write_pid
 
       path = @options[:state]
+      permission = @options[:state_permission]
       return unless path
 
       require 'puma/state_file'
@@ -113,8 +120,9 @@ module Puma
       sf.pid = Process.pid
       sf.control_url = @options[:control_url]
       sf.control_auth_token = @options[:control_auth_token]
+      sf.running_from = File.expand_path('.')
 
-      sf.save path
+      sf.save path, permission
     end
 
     # Delete the configured pidfile
@@ -169,29 +177,34 @@ module Puma
 
       setup_signals
       set_process_title
+      integrate_with_systemd
       @runner.run
 
       case @status
       when :halt
         log "* Stopping immediately!"
+        @runner.stop_control
       when :run, :stop
         graceful_stop
       when :restart
         log "* Restarting..."
         ENV.replace(previous_env)
-        @runner.before_restart
+        @runner.stop_control
         restart!
       when :exit
         # nothing
       end
-      @binder.close_unix_paths
+      close_binder_listeners unless @status == :restart
     end
 
-    # Return which tcp port the launcher is using, if it's using TCP
-    def connected_port
-      @binder.connected_port
+    # Return all tcp ports the launcher may be using, TCP or SSL
+    # @!attribute [r] connected_ports
+    # @version 5.0.0
+    def connected_ports
+      @binder.connected_ports
     end
 
+    # @!attribute [r] restart_args
     def restart_args
       cmd = @options[:restart_cmd]
       if cmd
@@ -202,7 +215,25 @@ module Puma
     end
 
     def close_binder_listeners
+      @runner.close_control_listeners
       @binder.close_listeners
+      unless @status == :restart
+        log "=== puma shutdown: #{Time.now} ==="
+        log "- Goodbye!"
+      end
+    end
+
+    # @!attribute [r] thread_status
+    # @version 5.0.0
+    def thread_status
+      Thread.list.each do |thread|
+        name = "Thread: TID-#{thread.object_id.to_s(36)}"
+        name += " #{thread['label']}" if thread['label']
+        name += " #{thread.name}" if thread.respond_to?(:name) && thread.name
+        backtrace = thread.backtrace || ["<no backtrace available>"]
+
+        yield name, backtrace
+      end
     end
 
     private
@@ -212,11 +243,10 @@ module Puma
     def write_pid
       path = @options[:pidfile]
       return unless path
-
-      File.open(path, 'w') { |f| f.puts Process.pid }
-      cur = Process.pid
+      cur_pid = Process.pid
+      File.write path, cur_pid, mode: 'wb:UTF-8'
       at_exit do
-        delete_pidfile if cur == Process.pid
+        delete_pidfile if cur_pid == Process.pid
       end
     end
 
@@ -225,7 +255,8 @@ module Puma
     end
 
     def restart!
-      @config.run_hooks :on_restart, self
+      @events.fire_on_restart!
+      @config.run_hooks :on_restart, self, @events
 
       if Puma.jruby?
         close_binder_listeners
@@ -241,21 +272,20 @@ module Puma
       else
         argv = restart_args
         Dir.chdir(@restart_dir)
+        ENV.update(@binder.redirects_for_restart_env)
         argv += [@binder.redirects_for_restart]
         Kernel.exec(*argv)
       end
     end
 
-    def dependencies_and_files_to_require_after_prune
+    # @!attribute [r] files_to_require_after_prune
+    def files_to_require_after_prune
       puma = spec_for_gem("puma")
 
-      deps = puma.runtime_dependencies.map do |d|
-        "#{d.name}:#{spec_for_gem(d.name).version}"
-      end
-
-      [deps, require_paths_for_gem(puma) + extra_runtime_deps_directories]
+      require_paths_for_gem(puma) + extra_runtime_deps_directories
     end
 
+    # @!attribute [r] extra_runtime_deps_directories
     def extra_runtime_deps_directories
       Array(@options[:extra_runtime_dependencies]).map do |d_name|
         if (spec = spec_for_gem(d_name))
@@ -267,6 +297,7 @@ module Puma
       end.flatten.compact
     end
 
+    # @!attribute [r] puma_wild_location
     def puma_wild_location
       puma = spec_for_gem("puma")
       dirs = require_paths_for_gem(puma)
@@ -275,6 +306,7 @@ module Puma
     end
 
     def prune_bundler
+      return if ENV['PUMA_BUNDLER_PRUNED']
       return unless defined?(Bundler)
       require_rubygems_min_version!(Gem::Version.new("2.2"), "prune_bundler")
       unless puma_wild_location
@@ -282,20 +314,48 @@ module Puma
         return
       end
 
-      deps, dirs = dependencies_and_files_to_require_after_prune
+      dirs = files_to_require_after_prune
 
       log '* Pruning Bundler environment'
       home = ENV['GEM_HOME']
-      Bundler.with_clean_env do
+      bundle_gemfile = Bundler.original_env['BUNDLE_GEMFILE']
+      bundle_app_config = Bundler.original_env['BUNDLE_APP_CONFIG']
+      with_unbundled_env do
         ENV['GEM_HOME'] = home
+        ENV['BUNDLE_GEMFILE'] = bundle_gemfile
         ENV['PUMA_BUNDLER_PRUNED'] = '1'
-        args = [Gem.ruby, puma_wild_location, '-I', dirs.join(':'), deps.join(',')] + @original_argv
+        ENV["BUNDLE_APP_CONFIG"] = bundle_app_config
+        args = [Gem.ruby, puma_wild_location, '-I', dirs.join(':')] + @original_argv
         # Ruby 2.0+ defaults to true which breaks socket activation
         args += [{:close_others => false}]
         Kernel.exec(*args)
       end
     end
 
+    #
+    # Puma's systemd integration allows Puma to inform systemd:
+    #  1. when it has successfully started
+    #  2. when it is starting shutdown
+    #  3. periodically for a liveness check with a watchdog thread
+    #
+
+    def integrate_with_systemd
+      return unless ENV["NOTIFY_SOCKET"]
+
+      begin
+        require 'puma/systemd'
+      rescue LoadError
+        log "Systemd integration failed. It looks like you're trying to use systemd notify but don't have sd_notify gem installed"
+        return
+      end
+
+      log "* Enabling systemd notification integration"
+
+      systemd = Systemd.new(@events)
+      systemd.hook_events
+      systemd.start_watchdog
+    end
+
     def spec_for_gem(gem_name)
       Bundler.rubygems.loaded_specs(gem_name)
     end
@@ -318,30 +378,15 @@ module Puma
     end
 
     def graceful_stop
+      @events.fire_on_stopped!
       @runner.stop_blocked
-      log "=== puma shutdown: #{Time.now} ==="
-      log "- Goodbye!"
-    end
-
-    def log_thread_status
-      Thread.list.each do |thread|
-        log "Thread TID-#{thread.object_id.to_s(36)} #{thread['label']}"
-        logstr = "Thread: TID-#{thread.object_id.to_s(36)}"
-        logstr += " #{thread.name}" if thread.respond_to?(:name)
-        log logstr
-
-        if thread.backtrace
-          log thread.backtrace.join("\n")
-        else
-          log "<no backtrace available>"
-        end
-      end
     end
 
     def set_process_title
       Process.respond_to?(:setproctitle) ? Process.setproctitle(title) : $0 = title
     end
 
+    # @!attribute [r] title
     def title
       buffer  = "puma #{Puma::Const::VERSION} (#{@options[:binds].join(',')})"
       buffer += " [#{@options[:tag]}]" if @options[:tag] && !@options[:tag].empty?
@@ -353,6 +398,7 @@ module Puma
       ENV['RACK_ENV'] = environment
     end
 
+    # @!attribute [r] environment
     def environment
       @environment
     end
@@ -456,8 +502,13 @@ module Puma
       end
 
       begin
-        Signal.trap "SIGINFO" do
-          log_thread_status
+        unless Puma.jruby? # INFO in use by JVM already
+          Signal.trap "SIGINFO" do
+            thread_status do |name, backtrace|
+              @events.log name
+              @events.log backtrace.map { |bt| "  #{bt}" }
+            end
+          end
         end
       rescue Exception
         # Not going to log this one, as SIGINFO is *BSD only and would be pretty annoying
@@ -471,5 +522,24 @@ module Puma
       raise "#{feature} is not supported on your version of RubyGems. " \
               "You must have RubyGems #{min_version}+ to use this feature."
     end
+
+    # @version 5.0.0
+    def with_unbundled_env
+      bundler_ver = Gem::Version.new(Bundler::VERSION)
+      if bundler_ver < Gem::Version.new('2.1.0')
+        Bundler.with_clean_env { yield }
+      else
+        Bundler.with_unbundled_env { yield }
+      end
+    end
+
+    def log_config
+      log "Configuration:"
+
+      @config.final_options
+        .each { |config_key, value| log "- #{config_key}: #{value}" }
+
+      log "\n"
+    end
   end
 end
