@@ -2,7 +2,6 @@
 
 require 'puma/server'
 require 'puma/const'
-require 'puma/minissl/context_builder'
 
 module Puma
   # Generic class that is used by `Puma::Cluster` and `Puma::Single` to
@@ -16,10 +15,16 @@ module Puma
       @app = nil
       @control = nil
       @started_at = Time.now
+      @wakeup = nil
     end
 
-    def daemon?
-      @options[:daemon]
+    def wakeup!
+      return unless @wakeup
+
+      @wakeup.write "!" unless @wakeup.closed?
+
+    rescue SystemCallError, IOError
+      Puma::Util.purge_interrupt_queue
     end
 
     def development?
@@ -34,7 +39,8 @@ module Puma
       @events.log str
     end
 
-    def before_restart
+    # @version 5.0.0
+    def stop_control
       @control.stop(true) if @control
     end
 
@@ -52,42 +58,27 @@ module Puma
 
       require 'puma/app/status'
 
-      uri = URI.parse str
-
       if token = @options[:control_auth_token]
         token = nil if token.empty? || token == 'none'
       end
 
       app = Puma::App::Status.new @launcher, token
 
-      control = Puma::Server.new app, @launcher.events
-      control.min_threads = 0
-      control.max_threads = 1
+      control = Puma::Server.new app, @launcher.events,
+        { min_threads: 0, max_threads: 1, queue_requests: false }
 
-      case uri.scheme
-      when "ssl"
-          log "* Starting control server on #{str}"
-          params = Util.parse_query uri.query
-          ctx = MiniSSL::ContextBuilder.new(params, @events).context
+      control.binder.parse [str], self, 'Starting control server'
 
-          control.add_ssl_listener uri.host, uri.port, ctx
-      when "tcp"
-        log "* Starting control server on #{str}"
-        control.add_tcp_listener uri.host, uri.port
-      when "unix"
-        log "* Starting control server on #{str}"
-        path = "#{uri.host}#{uri.path}"
-        mask = @options[:control_url_umask]
-
-        control.add_unix_listener path, mask
-      else
-        error "Invalid control URI: #{str}"
-      end
-
-      control.run
+      control.run thread_name: 'control'
       @control = control
     end
 
+    # @version 5.0.0
+    def close_control_listeners
+      @control.binder.close_listeners if @control
+    end
+
+    # @!attribute [r] ruby_engine
     def ruby_engine
       if !defined?(RUBY_ENGINE) || RUBY_ENGINE == "ruby"
         "ruby #{RUBY_VERSION}-p#{RUBY_PATCHLEVEL}"
@@ -105,12 +96,15 @@ module Puma
       max_t = @options[:max_threads]
 
       log "Puma starting in #{mode} mode..."
-      log "* Version #{Puma::Const::PUMA_VERSION} (#{ruby_engine}), codename: #{Puma::Const::CODE_NAME}"
-      log "* Min threads: #{min_t}, max threads: #{max_t}"
-      log "* Environment: #{ENV['RACK_ENV']}"
+      log "* Puma version: #{Puma::Const::PUMA_VERSION} (#{ruby_engine}) (\"#{Puma::Const::CODE_NAME}\")"
+      log "*  Min threads: #{min_t}"
+      log "*  Max threads: #{max_t}"
+      log "*  Environment: #{ENV['RACK_ENV']}"
 
-      if @options[:mode] == :tcp
-        log "* Mode: Lopez Express (tcp)"
+      if mode == "cluster"
+        log "*   Master PID: #{Process.pid}"
+      else
+        log "*          PID: #{Process.pid}"
       end
     end
 
@@ -124,23 +118,24 @@ module Puma
       append = @options[:redirect_append]
 
       if stdout
-        unless Dir.exist?(File.dirname(stdout))
-          raise "Cannot redirect STDOUT to #{stdout}"
-        end
+        ensure_output_directory_exists(stdout, 'STDOUT')
 
         STDOUT.reopen stdout, (append ? "a" : "w")
-        STDOUT.sync = true
         STDOUT.puts "=== puma startup: #{Time.now} ==="
+        STDOUT.flush unless STDOUT.sync
       end
 
       if stderr
-        unless Dir.exist?(File.dirname(stderr))
-          raise "Cannot redirect STDERR to #{stderr}"
-        end
+        ensure_output_directory_exists(stderr, 'STDERR')
 
         STDERR.reopen stderr, (append ? "a" : "w")
-        STDERR.sync = true
         STDERR.puts "=== puma startup: #{Time.now} ==="
+        STDERR.flush unless STDERR.sync
+      end
+
+      if @options[:mutate_stdout_and_stderr_to_sync_on_write]
+        STDOUT.sync = true
+        STDERR.sync = true
       end
     end
 
@@ -150,7 +145,6 @@ module Puma
         exit 1
       end
 
-      # Load the app before we daemonize.
       begin
         @app = @launcher.config.app
       rescue Exception => e
@@ -161,32 +155,22 @@ module Puma
       @launcher.binder.parse @options[:binds], self
     end
 
+    # @!attribute [r] app
     def app
       @app ||= @launcher.config.app
     end
 
     def start_server
-      min_t = @options[:min_threads]
-      max_t = @options[:max_threads]
-
       server = Puma::Server.new app, @launcher.events, @options
-      server.min_threads = min_t
-      server.max_threads = max_t
       server.inherit_binder @launcher.binder
-
-      if @options[:mode] == :tcp
-        server.tcp_mode!
-      end
-
-      if @options[:early_hints]
-        server.early_hints = true
-      end
-
-      unless development? || test?
-        server.leak_stack_on_error = false
-      end
-
       server
     end
+
+    private
+    def ensure_output_directory_exists(path, io_name)
+      unless Dir.exist?(File.dirname(path))
+        raise "Cannot redirect #{io_name} to #{path}"
+      end
+    end
   end
 end
