@@ -10,6 +10,13 @@ module Puma
       IO.pipe
     end
 
+    # An instance method on Thread has been provided to address https://bugs.ruby-lang.org/issues/13632,
+    # which currently effects some older versions of Ruby: 2.2.7 2.2.8 2.2.9 2.2.10 2.3.4 2.4.1
+    # Additional context: https://github.com/puma/puma/pull/1345
+    def purge_interrupt_queue
+      Thread.current.purge_interrupt_queue if Thread.current.respond_to? :purge_interrupt_queue
+    end
+
     # Unescapes a URI escaped string with +encoding+. +encoding+ will be the
     # target encoding of the string returned, and it defaults to UTF-8
     if defined?(::Encoding)
@@ -23,6 +30,17 @@ module Puma
     end
     module_function :unescape
 
+    # @version 5.0.0
+    def nakayoshi_gc(events)
+      events.log "! Promoting existing objects to old generation..."
+      4.times { GC.start(full_mark: false) }
+      if GC.respond_to?(:compact)
+        events.log "! Compacting..."
+        GC.compact
+      end
+      events.log "! Friendly fork preparation complete."
+    end
+
     DEFAULT_SEP = /[&;] */n
 
     # Stolen from Mongrel, with some small modifications:
@@ -50,7 +68,7 @@ module Puma
         end
       end
 
-      return params
+      params
     end
 
     # A case-insensitive Hash that preserves the original case of a
@@ -72,6 +90,7 @@ module Puma
         end
       end
 
+      # @!attribute [r] to_hash
       def to_hash
         hash = {}
         each { |k,v| hash[k] = v }
