@@ -1,18 +1,18 @@
 require_relative "helper"
-require_relative "helpers/ssl"
+require_relative "helpers/ssl" if ::Puma::HAS_SSL
+require_relative "helpers/tmp_path"
 
 require "puma/cli"
 require "json"
 
 class TestCLI < Minitest::Test
-  include SSLHelper
+  include SSLHelper if ::Puma::HAS_SSL
+  include TmpPath
 
   def setup
     @environment = 'production'
-    @tmp_file = Tempfile.new("puma-test")
-    @tmp_path = @tmp_file.path
-    @tmp_file.close!
 
+    @tmp_path = tmp_path('puma-test')
     @tmp_path2 = "#{@tmp_path}2"
 
     File.unlink @tmp_path  if File.exist? @tmp_path
@@ -37,11 +37,10 @@ class TestCLI < Minitest::Test
   end
 
   def test_control_for_tcp
-    tcp  = UniquePort.call
     cntl = UniquePort.call
     url = "tcp://127.0.0.1:#{cntl}/"
 
-    cli = Puma::CLI.new ["-b", "tcp://127.0.0.1:#{tcp}",
+    cli = Puma::CLI.new ["-b", "tcp://127.0.0.1:0",
                          "--control-url", url,
                          "--control-token", "",
                          "test/rackup/lobster.ru"], @events
@@ -57,8 +56,10 @@ class TestCLI < Minitest::Test
     body = s.read
     s.close
 
-    assert_match(/{ "started_at": "\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z", "backlog": 0, "running": 0, "pool_capacity": 16, "max_threads": 16 }/, body.split(/\r?\n/).last)
-    assert_match(/{ "started_at": "\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z", "backlog": 0, "running": 0, "pool_capacity": 16, "max_threads": 16 }/, Puma.stats)
+    assert_equal Puma.stats_hash, JSON.parse(Puma.stats, symbolize_names: true)
+
+    dmt = Puma::Configuration.new.default_max_threads
+    assert_match(/{"started_at":"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z","backlog":0,"running":0,"pool_capacity":#{dmt},"max_threads":#{dmt},"requests_count":0}/, body.split(/\r?\n/).last)
 
   ensure
     cli.launcher.stop
@@ -66,13 +67,15 @@ class TestCLI < Minitest::Test
   end
 
   def test_control_for_ssl
-    app_port = UniquePort.call
+    skip_unless :ssl
+
+    require "net/http"
     control_port = UniquePort.call
     control_host = "127.0.0.1"
     control_url = "ssl://#{control_host}:#{control_port}?#{ssl_query}"
     token = "token"
 
-    cli = Puma::CLI.new ["-b", "tcp://127.0.0.1:#{app_port}",
+    cli = Puma::CLI.new ["-b", "tcp://127.0.0.1:0",
                          "--control-url", control_url,
                          "--control-token", token,
                          "test/rackup/lobster.ru"], @events
@@ -92,18 +95,18 @@ class TestCLI < Minitest::Test
       body = http.request(req).body
     end
 
-    expected_stats = /{ "started_at": "\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z", "backlog": 0, "running": 0, "pool_capacity": 16, "max_threads": 16 }/
+    dmt = Puma::Configuration.new.default_max_threads
+    expected_stats = /{"started_at":"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z","backlog":0,"running":0,"pool_capacity":#{dmt},"max_threads":#{dmt}/
     assert_match(expected_stats, body.split(/\r?\n/).last)
-    assert_match(expected_stats, Puma.stats)
 
   ensure
-    cli.launcher.stop
-    t.join
+    cli.launcher.stop if cli
+    t.join if t
   end
 
   def test_control_clustered
-    skip NO_FORK_MSG  unless HAS_FORK
-    skip UNIX_SKT_MSG unless UNIX_SKT_EXIST
+    skip_unless :fork
+    skip_unless :unix
     url = "unix://#{@tmp_path}"
 
     cli = Puma::CLI.new ["-b", "unix://#{@tmp_path2}",
@@ -130,17 +133,16 @@ class TestCLI < Minitest::Test
 
     assert_equal 2, status["workers"]
 
-    # wait until the first status ping has come through
-    sleep 6
     s = UNIXSocket.new @tmp_path
     s << "GET /stats HTTP/1.0\r\n\r\n"
     body = s.read
     s.close
 
-    assert_match(/\{ "started_at": "\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z", "workers": 2, "phase": 0, "booted_workers": 2, "old_workers": 0, "worker_status": \[\{ "started_at": "\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z", "pid": \d+, "index": 0, "phase": 0, "booted": true, "last_checkin": "[^"]+", "last_status": \{ "backlog":0, "running":2, "pool_capacity":2, "max_threads": 2 \} \},\{ "started_at": "\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z", "pid": \d+, "index": 1, "phase": 0, "booted": true, "last_checkin": "[^"]+", "last_status": \{ "backlog":0, "running":2, "pool_capacity":2, "max_threads": 2 \} \}\] \}/, body.split("\r\n").last)
+    assert_match(/\{"started_at":"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z","workers":2,"phase":0,"booted_workers":2,"old_workers":0,"worker_status":\[\{"started_at":"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z","pid":\d+,"index":0,"phase":0,"booted":true,"last_checkin":"[^"]+","last_status":\{"backlog":0,"running":2,"pool_capacity":2,"max_threads":2,"requests_count":0\}\},\{"started_at":"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z","pid":\d+,"index":1,"phase":0,"booted":true,"last_checkin":"[^"]+","last_status":\{"backlog":0,"running":2,"pool_capacity":2,"max_threads":2,"requests_count":0\}\}\]\}/, body.split("\r\n").last)
   ensure
     if UNIX_SKT_EXIST && HAS_FORK
       cli.launcher.stop
+      t.join
 
       done = nil
       until done
@@ -149,13 +151,12 @@ class TestCLI < Minitest::Test
         done = log[/ - Goodbye!/]
       end
 
-      t.join
       $debugging_hold = false
     end
   end
 
   def test_control
-    skip UNIX_SKT_MSG unless UNIX_SKT_EXIST
+    skip_unless :unix
     url = "unix://#{@tmp_path}"
 
     cli = Puma::CLI.new ["-b", "unix://#{@tmp_path2}",
@@ -172,7 +173,8 @@ class TestCLI < Minitest::Test
     body = s.read
     s.close
 
-    assert_match(/{ "started_at": "\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z", "backlog": 0, "running": 0, "pool_capacity": 16, "max_threads": 16 }/, body.split("\r\n").last)
+    dmt = Puma::Configuration.new.default_max_threads
+    assert_match(/{"started_at":"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z","backlog":0,"running":0,"pool_capacity":#{dmt},"max_threads":#{dmt},"requests_count":0}/, body.split("\r\n").last)
   ensure
     if UNIX_SKT_EXIST
       cli.launcher.stop
@@ -181,7 +183,7 @@ class TestCLI < Minitest::Test
   end
 
   def test_control_stop
-    skip UNIX_SKT_MSG unless UNIX_SKT_EXIST
+    skip_unless :unix
     url = "unix://#{@tmp_path}"
 
     cli = Puma::CLI.new ["-b", "unix://#{@tmp_path2}",
@@ -203,6 +205,72 @@ class TestCLI < Minitest::Test
     t.join if UNIX_SKT_EXIST
   end
 
+  def test_control_requests_count
+    tcp  = UniquePort.call
+    cntl = UniquePort.call
+    url = "tcp://127.0.0.1:#{cntl}/"
+
+    cli = Puma::CLI.new ["-b", "tcp://127.0.0.1:#{tcp}",
+                         "--control-url", url,
+                         "--control-token", "",
+                         "test/rackup/lobster.ru"], @events
+
+    t = Thread.new do
+      cli.run
+    end
+
+    wait_booted
+
+    s = TCPSocket.new "127.0.0.1", cntl
+    s << "GET /stats HTTP/1.0\r\n\r\n"
+    body = s.read
+    s.close
+
+    assert_equal 0, JSON.parse(body.split(/\r?\n/).last)['requests_count']
+
+    # send real requests to server
+    3.times do
+      s = TCPSocket.new "127.0.0.1", tcp
+      s << "GET / HTTP/1.0\r\n\r\n"
+      body = s.read
+      s.close
+    end
+
+    s = TCPSocket.new "127.0.0.1", cntl
+    s << "GET /stats HTTP/1.0\r\n\r\n"
+    body = s.read
+    s.close
+
+    assert_equal 3, JSON.parse(body.split(/\r?\n/).last)['requests_count']
+  ensure
+    cli.launcher.stop
+    t.join
+  end
+
+  def test_control_thread_backtraces
+    skip_unless :unix
+    url = "unix://#{@tmp_path}"
+
+    cli = Puma::CLI.new ["-b", "unix://#{@tmp_path2}",
+                         "--control-url", url,
+                         "--control-token", "",
+                         "test/rackup/lobster.ru"], @events
+
+    t = Thread.new { cli.run }
+
+    wait_booted
+
+    s = UNIXSocket.new @tmp_path
+    s << "GET /thread-backtraces HTTP/1.0\r\n\r\n"
+    body = s.read
+    s.close
+
+    assert_match %r{Thread: TID-}, body.split("\r\n").last
+  ensure
+    cli.launcher.stop if cli
+    t.join if UNIX_SKT_EXIST
+  end
+
   def control_gc_stats(uri, cntl)
     cli = Puma::CLI.new ["-b", uri,
                          "--control-url", cntl,
@@ -254,7 +322,6 @@ class TestCLI < Minitest::Test
   end
 
   def test_control_gc_stats_tcp
-    skip_on :jruby, suffix: " - Hitting /gc route does not increment count"
     uri  = "tcp://127.0.0.1:#{UniquePort.call}/"
     cntl_port = UniquePort.call
     cntl = "tcp://127.0.0.1:#{cntl_port}/"
@@ -263,8 +330,7 @@ class TestCLI < Minitest::Test
   end
 
   def test_control_gc_stats_unix
-    skip_on :jruby, suffix: " - Hitting /gc route does not increment count"
-    skip UNIX_SKT_MSG unless UNIX_SKT_EXIST
+    skip_unless :unix
 
     uri  = "unix://#{@tmp_path2}"
     cntl = "unix://#{@tmp_path}"
@@ -273,7 +339,7 @@ class TestCLI < Minitest::Test
   end
 
   def test_tmp_control
-    skip_on :jruby, suffix: " - Unknown issue"
+    skip_if :jruby, suffix: " - Unknown issue"
 
     cli = Puma::CLI.new ["--state", @tmp_path, "--control-url", "auto"]
     cli.launcher.write_state
@@ -290,7 +356,7 @@ class TestCLI < Minitest::Test
   end
 
   def test_state_file_callback_filtering
-    skip NO_FORK_MSG unless HAS_FORK
+    skip_unless :fork
     cli = Puma::CLI.new [ "--config", "test/config/state_file_testing_config.rb",
                           "--state", @tmp_path ]
     cli.launcher.write_state
@@ -307,7 +373,7 @@ class TestCLI < Minitest::Test
   end
 
   def test_log_formatter_default_clustered
-    skip NO_FORK_MSG unless HAS_FORK
+    skip_unless :fork
 
     cli = Puma::CLI.new [ "-w 2" ]
     assert_instance_of Puma::Events::PidFormatter, cli.launcher.events.formatter
@@ -320,7 +386,7 @@ class TestCLI < Minitest::Test
   end
 
   def test_log_formatter_custom_clustered
-    skip NO_FORK_MSG unless HAS_FORK
+    skip_unless :fork
 
     cli = Puma::CLI.new [ "--config", "test/config/custom_log_formatter.rb", "-w 2" ]
     assert_instance_of Proc, cli.launcher.events.formatter
@@ -360,11 +426,38 @@ class TestCLI < Minitest::Test
     assert_equal %w[a b], extra_dependencies
   end
 
-  def test_environment
+  def test_environment_app_env
+    ENV['RACK_ENV'] = @environment
+    ENV['RAILS_ENV'] = @environment
+    ENV['APP_ENV'] = 'test'
+
+    cli = Puma::CLI.new []
+    cli.send(:setup_options)
+
+    assert_equal 'test', cli.instance_variable_get(:@conf).environment.call
+  ensure
+    ENV.delete 'APP_ENV'
+    ENV.delete 'RAILS_ENV'
+  end
+
+  def test_environment_rack_env
+    ENV['RACK_ENV'] = @environment
+
+    cli = Puma::CLI.new []
+    cli.send(:setup_options)
+
+    assert_equal @environment, cli.instance_variable_get(:@conf).environment.call
+  end
+
+  def test_environment_rails_env
     ENV.delete 'RACK_ENV'
+    ENV['RAILS_ENV'] = @environment
 
-    Puma::CLI.new ["--environment", @environment]
+    cli = Puma::CLI.new []
+    cli.send(:setup_options)
 
-    assert_equal ENV['RACK_ENV'], @environment
+    assert_equal @environment, cli.instance_variable_get(:@conf).environment.call
+  ensure
+    ENV.delete 'RAILS_ENV'
   end
 end
