@@ -1,3 +1,4 @@
+require 'puma/events'
 require_relative "helper"
 
 class TestEvents < Minitest::Test
@@ -23,6 +24,15 @@ class TestEvents < Minitest::Test
     assert_equal STDERR, events.stderr
   end
 
+  def test_stdio_respects_sync
+    events = Puma::Events.stdio
+
+    assert_equal STDOUT.sync, events.stdout.sync
+    assert_equal STDERR.sync, events.stderr.sync
+    assert_equal STDOUT, events.stdout
+    assert_equal STDERR, events.stderr
+  end
+
   def test_register_callback_with_block
     res = false
 
@@ -87,6 +97,14 @@ class TestEvents < Minitest::Test
     assert_equal "ready\n", out
   end
 
+  def test_null_log_does_nothing
+    out, _ = capture_io do
+      Puma::Events.null.log("ready")
+    end
+
+    assert_equal "", out
+  end
+
   def test_write_writes_to_stdout
     out, _ = capture_io do
       Puma::Events.stdio.write("ready")
@@ -119,14 +137,16 @@ class TestEvents < Minitest::Test
     did_exit = false
 
     _, err = capture_io do
-      Puma::Events.stdio.error("interrupted")
+      begin
+        Puma::Events.stdio.error("interrupted")
+      rescue SystemExit
+        did_exit = true
+      ensure
+        assert did_exit
+      end
     end
 
-    assert_equal "ERROR: interrupted", err
-  rescue SystemExit
-    did_exit = true
-  ensure
-    assert did_exit
+    assert_match %r!ERROR: interrupted!, err
   end
 
   def test_pid_formatter
@@ -164,17 +184,55 @@ class TestEvents < Minitest::Test
     events = Puma::Events.strings
     server = Puma::Server.new app, events
 
-    server.add_tcp_listener host, port
+    port = (server.add_tcp_listener host, 0).addr[1]
     server.run
 
-    sock = TCPSocket.new host, server.connected_port
+    sock = TCPSocket.new host, port
     path = "/"
     params = "a"*1024*10
 
     sock << "GET #{path}?a=#{params} HTTP/1.1\r\nConnection: close\r\n\r\n"
     sock.read
     sleep 0.1 # important so that the previous data is sent as a packet
-    assert_match %r!HTTP parse error, malformed request \(#{path}\)!, events.stderr.string
-    server.stop(true)
+    assert_match %r!HTTP parse error, malformed request!, events.stderr.string
+    assert_match %r!\("GET #{path}" - \(-\)\)!, events.stderr.string
+  ensure
+    sock.close if sock && !sock.closed?
+    server.stop true
   end
+
+  # test_puma_server_ssl.rb checks that ssl errors are raised correctly,
+  # but it mocks the actual error code.  This test the code, but it will
+  # break if the logged message changes
+  def test_ssl_error
+    events = Puma::Events.strings
+
+    ssl_mock = -> (addr, subj) {
+      obj = Object.new
+      obj.define_singleton_method(:peeraddr) { addr }
+      if subj
+        cert = Object.new
+        cert.define_singleton_method(:subject) { subj }
+        obj.define_singleton_method(:peercert) { cert }
+      else
+        obj.define_singleton_method(:peercert) { nil }
+      end
+      obj
+    }
+
+    events.ssl_error OpenSSL::SSL::SSLError, ssl_mock.call(['127.0.0.1'], 'test_cert')
+    error = events.stderr.string
+    assert_includes error, "SSL error"
+    assert_includes error, "peer: 127.0.0.1"
+    assert_includes error, "cert: test_cert"
+
+    events.stderr.string = ''
+
+    events.ssl_error OpenSSL::SSL::SSLError, ssl_mock.call(nil, nil)
+    error = events.stderr.string
+    assert_includes error, "SSL error"
+    assert_includes error, "peer: <unknown>"
+    assert_includes error, "cert: :"
+
+  end if ::Puma::HAS_SSL
 end
