@@ -114,7 +114,7 @@ class Http11ParserTest < Minitest::Test
   end
 
   def test_semicolon_in_path
-    skip_on :jruby # Not yet supported on JRuby, see https://github.com/puma/puma/issues/1978
+    skip_if :jruby # Not yet supported on JRuby, see https://github.com/puma/puma/issues/1978
     parser = Puma::HttpParser.new
     req = {}
     get = "GET /forums/1/path;stillpath/2375?page=1 HTTP/1.1\r\n\r\n"
@@ -137,22 +137,22 @@ class Http11ParserTest < Minitest::Test
       res << Digest::SHA1.digest(rand(count * 100).to_s) * (count / 20)
     end
 
-    return res
+    res
   end
 
   def test_max_uri_path_length
     parser = Puma::HttpParser.new
     req = {}
 
-    # Support URI path length to a max of 2048
-    path = "/" + rand_data(1000, 100)
+    # Support URI path length to a max of 8192
+    path = "/" + rand_data(7000, 100)
     http = "GET #{path} HTTP/1.1\r\n\r\n"
     parser.execute(req, http, 0)
     assert_equal path, req['REQUEST_PATH']
     parser.reset
 
-    # Raise exception if URI path length > 2048
-    path = "/" + rand_data(3000, 100)
+    # Raise exception if URI path length > 8192
+    path = "/" + rand_data(9000, 100)
     http = "GET #{path} HTTP/1.1\r\n\r\n"
     assert_raises Puma::HttpParserError do
       parser.execute(req, http, 0)
@@ -208,4 +208,34 @@ class Http11ParserTest < Minitest::Test
 
     assert_equal "Strip This", req["HTTP_X_STRIP_ME"]
   end
+
+  def test_newline_smuggler
+    parser = Puma::HttpParser.new
+    req = {}
+    http = "GET / HTTP/1.1\r\nHost: localhost:8080\r\nDummy: x\nDummy2: y\r\n\r\n"
+
+    parser.execute(req, http, 0) rescue nil # We test the raise elsewhere.
+
+    assert parser.error?, "Parser SHOULD have error"
+  end
+
+  def test_newline_smuggler_two
+    parser = Puma::HttpParser.new
+    req = {}
+    http = "GET / HTTP/1.1\r\nHost: localhost:8080\r\nDummy: x\r\nDummy: y\nDummy2: z\r\n\r\n"
+
+    parser.execute(req, http, 0) rescue nil
+
+    assert parser.error?, "Parser SHOULD have error"
+  end
+
+  def test_htab_in_header_val
+    parser = Puma::HttpParser.new
+    req = {}
+    http = "GET / HTTP/1.1\r\nHost: localhost:8080\r\nDummy: Valid\tValue\r\n\r\n"
+
+    parser.execute(req, http, 0)
+
+    assert_equal "Valid\tValue", req['HTTP_DUMMY']
+  end
 end
