@@ -21,6 +21,10 @@ class TestNullIO < Minitest::Test
     assert_nil nio.gets
   end
 
+  def test_string_returns_empty_string
+    assert_equal "", nio.string
+  end
+
   def test_each_never_yields
     nio.instance_variable_set(:@foo, :baz)
     nio.each { @foo = :bar }
@@ -52,4 +56,12 @@ class TestNullIO < Minitest::Test
   def test_size
     assert_equal 0, nio.size
   end
+
+  def test_sync_returns_true
+    assert_equal true, nio.sync
+  end
+
+  def test_flush_returns_self
+    assert_equal nio, nio.flush
+  end
 end
