@@ -1,36 +1,28 @@
+# Nothing in this file runs if Puma isn't compiled with ssl support
+#
+# helper is required first since it loads Puma, which needs to be
+# loaded so HAS_SSL is defined
 require_relative "helper"
-require "puma/minissl"
-require "puma/puma_http11"
 
-#———————————————————————————————————————————————————————————————————————————————
-#             NOTE: ALL TESTS BYPASSED IF DISABLE_SSL IS TRUE
-#———————————————————————————————————————————————————————————————————————————————
+if ::Puma::HAS_SSL
+  require "puma/minissl"
+  require "net/http"
 
-class SSLEventsHelper < ::Puma::Events
-  attr_accessor :addr, :cert, :error
-
-  def ssl_error(server, peeraddr, peercert, error)
-    self.addr = peeraddr
-    self.cert = peercert
-    self.error = error
+  # net/http (loaded in helper) does not necessarily load OpenSSL
+  require "openssl" unless Object.const_defined? :OpenSSL
+  if Puma::IS_JRUBY
+    puts "", RUBY_DESCRIPTION, "RUBYOPT: #{ENV['RUBYOPT']}",
+      "                         OpenSSL",
+      "OPENSSL_LIBRARY_VERSION: #{OpenSSL::OPENSSL_LIBRARY_VERSION}",
+      "        OPENSSL_VERSION: #{OpenSSL::OPENSSL_VERSION}", ""
+  else
+    puts "", RUBY_DESCRIPTION, "RUBYOPT: #{ENV['RUBYOPT']}",
+      "                         Puma::MiniSSL                   OpenSSL",
+      "OPENSSL_LIBRARY_VERSION: #{Puma::MiniSSL::OPENSSL_LIBRARY_VERSION.ljust 32}#{OpenSSL::OPENSSL_LIBRARY_VERSION}",
+      "        OPENSSL_VERSION: #{Puma::MiniSSL::OPENSSL_VERSION.ljust 32}#{OpenSSL::OPENSSL_VERSION}", ""
   end
 end
 
-DISABLE_SSL = begin
-              Puma::Server.class
-              Puma::MiniSSL.check
-              # net/http (loaded in helper) does not necessarily load OpenSSL
-              require "openssl" unless Object.const_defined? :OpenSSL
-              puts "", RUBY_DESCRIPTION,
-                   "                         Puma::MiniSSL                   OpenSSL",
-                   "OPENSSL_LIBRARY_VERSION: #{Puma::MiniSSL::OPENSSL_LIBRARY_VERSION.ljust 32}#{OpenSSL::OPENSSL_LIBRARY_VERSION}",
-                   "        OPENSSL_VERSION: #{Puma::MiniSSL::OPENSSL_VERSION.ljust 32}#{OpenSSL::OPENSSL_VERSION}", ""
-            rescue
-              true
-            else
-              false
-            end
-
 class TestPumaServerSSL < Minitest::Test
   parallelize_me!
   def setup
@@ -45,7 +37,6 @@ class TestPumaServerSSL < Minitest::Test
 
   # yields ctx to block, use for ctx setup & configuration
   def start_server
-    @port = UniquePort.call
     @host = "127.0.0.1"
 
     app = lambda { |env| [200, {}, [env['rack.url_scheme']]] }
@@ -66,7 +57,7 @@ class TestPumaServerSSL < Minitest::Test
 
     @events = SSLEventsHelper.new STDOUT, STDERR
     @server = Puma::Server.new app, @events
-    @ssl_listener = @server.add_ssl_listener @host, @port, ctx
+    @port = (@server.add_ssl_listener @host, 0, ctx).addr[1]
     @server.run
 
     @http = Net::HTTP.new @host, @port
@@ -93,8 +84,10 @@ class TestPumaServerSSL < Minitest::Test
     # Open a connection and give enough data to trigger a read, then wait
     ctx = OpenSSL::SSL::SSLContext.new
     ctx.verify_mode = OpenSSL::SSL::VERIFY_NONE
-    socket = OpenSSL::SSL::SSLSocket.new TCPSocket.new(@host, @port), ctx
-    socket.write "x"
+    port = @server.connected_ports[0]
+    socket = OpenSSL::SSL::SSLSocket.new TCPSocket.new(@host, port), ctx
+    socket.connect
+    socket.write "HEAD"
     sleep 0.1
 
     # Capture the amount of threads being used after connecting and being idle
@@ -163,7 +156,7 @@ class TestPumaServerSSL < Minitest::Test
     skip("TLSv1 protocol is unavailable") if Puma::MiniSSL::OPENSSL_NO_TLS1
     start_server { |ctx| ctx.no_tlsv1 = true }
 
-    if @http.respond_to? :max_version=
+    if OpenSSL::SSL::SSLContext.private_instance_methods(false).include?(:set_minmax_proto_version)
       @http.max_version = :TLS1
     else
       @http.ssl_version = :TLSv1
@@ -183,7 +176,7 @@ class TestPumaServerSSL < Minitest::Test
   def test_tls_v1_1_rejection
     start_server { |ctx| ctx.no_tlsv1_1 = true }
 
-    if @http.respond_to? :max_version=
+    if OpenSSL::SSL::SSLContext.private_instance_methods(false).include?(:set_minmax_proto_version)
       @http.max_version = :TLS1_1
     else
       @http.ssl_version = :TLSv1_1
@@ -199,23 +192,57 @@ class TestPumaServerSSL < Minitest::Test
       assert_match(msg, @events.error.message) if @events.error
     end
   end
-end unless DISABLE_SSL
+
+  def test_http_rejection
+    body_http  = nil
+    body_https = nil
+
+    start_server
+
+    http = Net::HTTP.new @host, @server.connected_ports[0]
+    http.use_ssl = false
+    http.read_timeout = 6
+
+    tcp = Thread.new do
+      req_http = Net::HTTP::Get.new "/", {}
+      # Net::ReadTimeout - TruffleRuby
+      assert_raises(Errno::ECONNREFUSED, EOFError, Net::ReadTimeout, Net::OpenTimeout) do
+        http.start.request(req_http) { |rep| body_http = rep.body }
+      end
+    end
+
+    ssl = Thread.new do
+      @http.start do
+        req_https = Net::HTTP::Get.new "/", {}
+        @http.request(req_https) { |rep_https| body_https = rep_https.body }
+      end
+    end
+
+    tcp.join
+    ssl.join
+    http.finish
+    sleep 1.0
+
+    assert_nil body_http
+    assert_equal "https", body_https
+
+    thread_pool = @server.instance_variable_get(:@thread_pool)
+    busy_threads = thread_pool.spawned - thread_pool.waiting
+
+    assert busy_threads.zero?, "Our connection is wasn't dropped"
+  end
+end if ::Puma::HAS_SSL
 
 # client-side TLS authentication tests
 class TestPumaServerSSLClient < Minitest::Test
-  parallelize_me!
+  parallelize_me! unless ::Puma.jruby?
 
   CERT_PATH = File.expand_path "../examples/puma/client-certs", __dir__
 
-  def assert_ssl_client_error_match(error, subject=nil, &blk)
-    host = "localhost"
-    port = UniquePort.call
-
-    app = lambda { |env| [200, {}, [env['rack.url_scheme']]] }
-
-    ctx = Puma::MiniSSL::Context.new
+  # Context can be shared, may help with JRuby
+  CTX = Puma::MiniSSL::Context.new.tap { |ctx|
     if Puma.jruby?
-      ctx.keystore = "#{CERT_PATH}/keystore.jks"
+      ctx.keystore =  "#{CERT_PATH}/keystore.jks"
       ctx.keystore_pass = 'jruby_puma'
     else
       ctx.key  = "#{CERT_PATH}/server.key"
@@ -223,14 +250,21 @@ class TestPumaServerSSLClient < Minitest::Test
       ctx.ca   = "#{CERT_PATH}/ca.crt"
     end
     ctx.verify_mode = Puma::MiniSSL::VERIFY_PEER | Puma::MiniSSL::VERIFY_FAIL_IF_NO_PEER_CERT
+  }
+
+  def assert_ssl_client_error_match(error, subject=nil, &blk)
+    host = "localhost"
+    port = 0
+
+    app = lambda { |env| [200, {}, [env['rack.url_scheme']]] }
 
     events = SSLEventsHelper.new STDOUT, STDERR
     server = Puma::Server.new app, events
-    server.add_ssl_listener host, port, ctx
+    server.add_ssl_listener host, port, CTX
     host_addrs = server.binder.ios.map { |io| io.to_io.addr[2] }
     server.run
 
-    http = Net::HTTP.new host, port
+    http = Net::HTTP.new host, server.connected_ports[0]
     http.use_ssl = true
     http.verify_mode = OpenSSL::SSL::VERIFY_NONE
 
@@ -242,7 +276,8 @@ class TestPumaServerSSLClient < Minitest::Test
         req = Net::HTTP::Get.new "/", {}
         http.request(req)
       end
-    rescue OpenSSL::SSL::SSLError, EOFError
+    rescue OpenSSL::SSL::SSLError, EOFError, Errno::ECONNRESET
+      # Errno::ECONNRESET TruffleRuby
       client_error = true
       # closes socket if open, may not close on error
       http.send :do_finish
@@ -258,7 +293,7 @@ class TestPumaServerSSLClient < Minitest::Test
       assert_equal subject, events.cert.subject.to_s if subject
     end
   ensure
-    server.stop(true)
+    server.stop(true) if server
   end
 
   def test_verify_fail_if_no_client_cert
@@ -297,4 +332,4 @@ class TestPumaServerSSLClient < Minitest::Test
       http.verify_mode = OpenSSL::SSL::VERIFY_PEER
     end
   end
-end unless DISABLE_SSL
+end if ::Puma::HAS_SSL
